/*
 * XetPan: a libEtPan! based Mail User Agent
 * Copyright (C) 2002-2003 Melvin Hadasht
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>

#include "intl.h"

#include "logging.h"
#include "defines.h"
#include "ui.h"
#include "remote.h"

/* TODO: make the socket more portable */
#define XETPAN_LOCK_SOCKET    ".xetpan.socket"

#define REMOTE_MSG_PREFIX     "XetPan>"
#define REMOTE_MSG_PREFIX_LEN strlen(REMOTE_MSG_PREFIX)
#define REMOTE_SIZE_TYPE      unsigned short int
#define REMOTE_SIZE_LEN       sizeof(REMOTE_SIZE_TYPE)
#define REMOTE_SIZE_MAX       0xFFFF
#define REMOTE_TOKEN_LEN      1
#define REMOTE_MSG_BUF        (xetpan->remote_msg[xetpan->remote_idx])
#define REMOTE_PREAMBLE_LEN   (REMOTE_MSG_PREFIX_LEN + REMOTE_SIZE_LEN + \
                              REMOTE_TOKEN_LEN)

enum 
  {
    REMOTE_READING_MSG_PREFIX,
    REMOTE_READING_MSG_SIZE,
    REMOTE_READING_MSG_CONTENT,
  };

enum
  {
    REMOTE_STATE_MSG_COMPLETED,
    REMOTE_STATE_MSG_INCOMPLETE,
    REMOTE_STATE_MSG_ERROR,
    REMOTE_STATE_MSG_ABORTED
  };

/* Remote socket functions*/
static void
Remote_create_sockaddr(char *lock_socket, struct sockaddr_un *serv_addr,
		       int *servlen);

static void
Remote_clear_msg_buffer(XePRemote *remote);

/* Remote Messages basic I/O functions used by server and client */
static int
Remote_send_msg_chunk(XePRemote *remote);

static int
Remote_receive_msg_chunk(XePRemote *remote);

static void
Remote_send_msg_cb(int fd, short int condition, void *data);

static void
Remote_receive_msg_cb(int fd, short int condition, void *data);

/* Server side functions */
static void
Remote_close_client_connection(XePRemote *remote);

static int
Remote_execute_cmd(XePRemote *remote);

/*************************************************************/

/* Remote socket handling */
static void
Remote_create_sockaddr(char * lock_socket, 
		       struct sockaddr_un *serv_addr,
		       int *servlen)
{
  char *home;
  memset((char *)serv_addr, 0, sizeof(*serv_addr));
  serv_addr->sun_family = AF_UNIX;

  /* TODO: make the socket code more portable */
  home = getenv("HOME");
  if (home == NULL)
    home = "/tmp";
  else if (strlen(home) + strlen(XETPAN_LOCK_SOCKET) + 2 >= UNIX_PATH_MAX)
    home = "/tmp";
  sprintf(lock_socket, "%s/" XETPAN_LOCK_SOCKET, home);
  
  lock_socket[UNIX_PATH_MAX - 1] = 0x00;

  strcpy(serv_addr->sun_path, lock_socket);

  *servlen = strlen(serv_addr->sun_path) + sizeof(serv_addr->sun_family);
}

/* Remote Message building functions */
int
a_Remote_build_msg(XePRemote *remote, char cmd, const char *option)
{
  const char *arg = "";
  int ret;
  REMOTE_SIZE_TYPE cmd_len;
  
  if (option)
    arg = option;

  A_LOGGING_DEBUG("Creating msg '%c'. Option %s", cmd, arg);

  if (REMOTE_PREAMBLE_LEN + strlen(arg) > REMOTE_SIZE_MAX)
    {
      A_LOGGING_ERROR("Remote message too long. Aborting.");
      ret = XETPAN_ERR_MSG_TOO_LONG;
    }
  else
    {
      cmd_len = (REMOTE_SIZE_TYPE) (REMOTE_PREAMBLE_LEN + strlen(arg));
      
      remote->msg = (char *) realloc(remote->msg, cmd_len + 1);
      remote->msg[cmd_len] = 0x00;
      remote->msg_size = cmd_len + 1;
      remote->msg_len = cmd_len;
      
      sprintf(remote->msg, "%s", REMOTE_MSG_PREFIX);
      remote->msg[REMOTE_MSG_PREFIX_LEN]     = cmd_len &  0x00ff;
      remote->msg[REMOTE_MSG_PREFIX_LEN + 1] = cmd_len /  0x0100;
      remote->msg[REMOTE_MSG_PREFIX_LEN + REMOTE_SIZE_LEN] = cmd;
      
      memcpy(remote->msg + REMOTE_PREAMBLE_LEN, arg, strlen(arg));
      
      remote->msg_idx = 0;
      
      ret = XETPAN_NO_ERR;
    }
  
  return ret;
}

static void
Remote_clear_msg_buffer(XePRemote *remote)
{
  remote->msg_len = 0;
  remote->msg_idx = 0;
}

static void
Remote_close_client_connection(XePRemote *remote)
{
  A_LOGGING_DEBUG("Closing remote client fd %d", remote->fd);
  close(remote->fd);
  if (remote->msg)
    free(remote->msg);
  free(remote);
}

/* Remote Messages basic I/O functions used by server and client */
static int
Remote_send_msg_chunk(XePRemote *remote)
{
  int state;

  state = REMOTE_STATE_MSG_INCOMPLETE;

  if (remote->msg_idx <= remote->msg_len) 
    {
      ssize_t count;
      count = write(remote->fd, remote->msg + remote->msg_idx,
		    remote->msg_len - remote->msg_idx);
      if (count < 0)
	{
	  A_LOGGING_SYSTEM_ERROR("write: %s", strerror(errno));
	  A_LOGGING_ERROR(_("Command ignored"));
	  state = REMOTE_STATE_MSG_ERROR;
	}
      else
	{
	  remote->msg_idx += count;
	  if (remote->msg_idx == remote->msg_len)
	    {
	      remote->msg_idx = 0;
	      state = REMOTE_STATE_MSG_COMPLETED;
	    }
	}
    }
  
  return state;
}

static int
Remote_receive_msg_chunk(XePRemote *remote)
{
  ssize_t count;
  int state;
  int reading;

  if (remote->msg_idx < REMOTE_MSG_PREFIX_LEN)
    reading = REMOTE_READING_MSG_PREFIX;
  else if (remote->msg_idx < REMOTE_MSG_PREFIX_LEN + REMOTE_SIZE_LEN)
    reading = REMOTE_READING_MSG_SIZE;
  else 
    reading = REMOTE_READING_MSG_CONTENT;

  switch (reading)
    {
    case REMOTE_READING_MSG_PREFIX:
      count = read(remote->fd, 
		   remote->msg + remote->msg_idx, 1);
      if (count < 0)
	state = REMOTE_STATE_MSG_ERROR;
      else if (count == 0)
	state = REMOTE_STATE_MSG_ABORTED;
      else 
	{
	  if (remote->msg[remote->msg_idx] == 
	      REMOTE_MSG_PREFIX[remote->msg_idx])
	    remote->msg_idx++;
	  else
	    remote->msg_idx = 0;
	  state = REMOTE_STATE_MSG_INCOMPLETE;
	}
      break;
    case REMOTE_READING_MSG_SIZE:
      count = read(remote->fd,
		   remote->msg + remote->msg_idx, 
		   REMOTE_MSG_PREFIX_LEN + REMOTE_SIZE_LEN - remote->msg_idx);
      if (count < 0)
	state = REMOTE_STATE_MSG_ERROR;
      else if (count == 0)
	state = REMOTE_STATE_MSG_ABORTED;
      else
	{
	  remote->msg_idx += count;
	  state = REMOTE_STATE_MSG_INCOMPLETE;
	}
      break;
    case REMOTE_READING_MSG_CONTENT:
      remote->msg_len  = remote->msg[REMOTE_MSG_PREFIX_LEN];
      remote->msg_len |= remote->msg[REMOTE_MSG_PREFIX_LEN + 1] *
	0x0100;
      if (remote->msg_len + 1 > remote->msg_size)
	{
	  char *msg;
	  msg = (char *) realloc(remote->msg, remote->msg_len + 1);

	  if (msg == NULL)
	    {
	      A_LOGGING_SYSTEM_ERROR("realloc: %s", strerror(errno));
	      state = REMOTE_STATE_MSG_ABORTED;
	      break;
	    }

	  remote->msg = msg;
	  remote->msg_size = remote->msg_len + 1;

	}
      if (remote->msg_idx < remote->msg_len)
	{
	  count = read(remote->fd, 
		       remote->msg + remote->msg_idx,
		       remote->msg_len - remote->msg_idx);
	  if (count < 0)
	    state = REMOTE_STATE_MSG_ERROR;
	  else if (count == 0)
	    state = REMOTE_STATE_MSG_ABORTED;
	  else
	    {
	      remote->msg_idx += count;
	      if (remote->msg_idx == remote->msg_len)
		{
		  remote->msg[remote->msg_len] = 0x00;
		  remote->msg_idx = 0;
		  state = REMOTE_STATE_MSG_COMPLETED;
		}
	      else
		state = REMOTE_STATE_MSG_INCOMPLETE;
	    }
	}
      else
	state = REMOTE_STATE_MSG_INCOMPLETE;
      break;
    default:
      state = REMOTE_STATE_MSG_INCOMPLETE;
    }

  if (state == REMOTE_STATE_MSG_ERROR)
    A_LOGGING_SYSTEM_ERROR("read: %s", strerror(errno));

  return state;
}

/* Server I/O callbacks for asynchroneous communication */
static void
Remote_send_msg_cb(int fd, short int condition, void *data)
{
  XePRemote *remote = (XePRemote *) data;
  int state;

  A_LOGGING_DEBUG("fd %d condition 0x%x", fd, condition);

  if (condition & (XETPAN_POLLHUP | XETPAN_POLLERR))
    {
      a_UI_remove_io_watch(remote->xetpan, remote->watch_id);
      Remote_close_client_connection(remote);
    }
  else
    {
      state = Remote_send_msg_chunk(remote);
      if (state == REMOTE_STATE_MSG_COMPLETED)
	{
	  a_UI_remove_io_watch(remote->xetpan, remote->watch_id);
	  Remote_clear_msg_buffer(remote);
	  remote->watch_id = a_UI_add_io_watch(remote->xetpan, 
					       remote->fd,
					       XETPAN_POLLIN |
					       XETPAN_POLLHUP |
					       XETPAN_POLLERR,
					       Remote_receive_msg_cb,
					       remote);
	}
      else if (state != REMOTE_STATE_MSG_INCOMPLETE)
	{
	  a_UI_remove_io_watch(remote->xetpan, remote->watch_id);
	  Remote_close_client_connection(remote);
	}
    }
}

static void
Remote_receive_msg_cb(int fd, short int condition, void *data)
{
  XePRemote *remote = (XePRemote *) data;
  int state;

  A_LOGGING_DEBUG("fd %d condition 0x%x",
		  fd, condition);

  if (condition & (XETPAN_POLLHUP | XETPAN_POLLERR))
    {
      a_UI_remove_io_watch(remote->xetpan, remote->watch_id);
      Remote_close_client_connection(remote);
      
    }
  else
    {
      state = Remote_receive_msg_chunk(remote);
      
      if (state == REMOTE_STATE_MSG_COMPLETED)
	{
	  a_UI_remove_io_watch(remote->xetpan, remote->watch_id);
	  state = Remote_execute_cmd(remote);
	  if (state != XETPAN_REMOTE_CLIENT)
	    {
	      if (state != XETPAN_NO_ERR)
		Remote_close_client_connection(remote);
	      else
	    a_Remote_send_cmd_result(remote);
	    }
	}
      else if (state == REMOTE_STATE_MSG_ABORTED || 
	       state == REMOTE_STATE_MSG_ERROR)
	{
	  a_UI_remove_io_watch(remote->xetpan, remote->watch_id);
	  Remote_close_client_connection(remote);
	}
    }
}

void
a_Remote_send_cmd_result(XePRemote *remote)
{
  remote->watch_id = a_UI_add_io_watch(remote->xetpan, 
				       remote->fd,
				       XETPAN_POLLOUT | XETPAN_POLLHUP |
				       XETPAN_POLLERR,
				       Remote_send_msg_cb,
				       remote);
}

static int
Remote_execute_cmd(XePRemote *remote)
{
  char cmd;
  char *option;
  int option_len;
  int ret;
  
  cmd = remote->msg[REMOTE_MSG_PREFIX_LEN + REMOTE_SIZE_LEN];
  option_len = remote->msg_len - REMOTE_PREAMBLE_LEN;
  option = (char *) malloc(option_len + 1);
  memcpy(option, remote->msg + REMOTE_PREAMBLE_LEN, option_len);
  option[option_len] = 0x00;

  Remote_clear_msg_buffer(remote);

  ret = a_XetPan_execute_remote_cmd(remote, cmd, option);

  free(option);

  return ret;
}

/* Client function */
void
a_Remote_send_command(XetPan *xetpan, char cmd, const char *option)
{
  int state;
  XePRemote *remote;

  remote = (XePRemote *) calloc(1, sizeof(XePRemote));
  if (remote == NULL)
    {
      A_LOGGING_SYSTEM_ERROR("calloc: %s", strerror(errno));
    }
  else if (a_Remote_build_msg(remote, cmd, option) == XETPAN_NO_ERR)
    {
      remote->xetpan = xetpan;
      remote->fd = xetpan->server_fd;
      while ((state = Remote_send_msg_chunk(remote)) == 
	     REMOTE_STATE_MSG_INCOMPLETE)
	{
	  /* Wait loop: do nothing */
	}
      
      if (state == REMOTE_STATE_MSG_COMPLETED)
	{
	  Remote_clear_msg_buffer(remote);
	  while((state = Remote_receive_msg_chunk(remote)) == 
		REMOTE_STATE_MSG_INCOMPLETE)
	    {
	      /* Wait loop: do nothing */
	    }
	  
	  if (state == REMOTE_STATE_MSG_COMPLETED)
	    {
	      printf("%s\n", remote->msg + REMOTE_PREAMBLE_LEN);
	    }
	}
      if (remote->msg)
	free(remote->msg);
      free(remote);
    }
}

/* Server functions */
static void
Remote_accept_from_socket(int server_fd, short int condition, void *data)
{
  XetPan *xetpan = (XetPan *) data;
  XePRemote *remote;

  struct sockaddr_un cli_addr;
  socklen_t clilen = sizeof(cli_addr);
  int fd;
  char *msg;

  fd = accept(xetpan->server_fd, (struct sockaddr *)&cli_addr,
	      &clilen);

  if (fd < 0)
    {
      A_LOGGING_SYSTEM_ERROR("accept: %s", strerror(errno));
    }
  else
    {
      remote = (XePRemote *) calloc(1, sizeof(XePRemote));
      if (remote == NULL)
	{
	  A_LOGGING_SYSTEM_ERROR("calloc: %s", strerror(errno));
	  close(fd);
	}
      else
	{
	  msg = (char *) calloc(REMOTE_MSG_SIZE, sizeof(char));
	  if (msg == NULL)
	    {
	      A_LOGGING_SYSTEM_ERROR("calloc: %s", strerror(errno));
	      close(fd);
	      free(remote);
	    }
	  else
	    {
	      A_LOGGING_INFO("Remote client connected");
	      A_LOGGING_DEBUG("remote client connected to fd %d", fd);

	      remote->xetpan = xetpan;
	      remote->fd = fd;
	      remote->msg = msg;
	      remote->msg_size = REMOTE_MSG_SIZE;
	      remote->watch_id = a_UI_add_io_watch(xetpan, fd, XETPAN_POLLIN |
						   XETPAN_POLLHUP | 
						   XETPAN_POLLERR,
						   Remote_receive_msg_cb,
						   remote);
	    }
	}
    }
}

int
a_Remote_be_a_server(XetPan *xetpan)
{
  int server_fd;
  unsigned int server_watchid;
  struct sockaddr_un serv_addr;
  int servlen;
  int ret;
  
  Remote_create_sockaddr(xetpan->lock_socket, &serv_addr, &servlen);
  
  server_fd = socket(AF_UNIX, SOCK_STREAM, 0);

  ret = FALSE;

  if (server_fd < 0)
    {
      A_LOGGING_SYSTEM_ERROR("socket: %s", strerror(errno));
      /* Maybe prevent the user? */
    }
  else
    {
      if (fcntl(server_fd, F_SETFL, O_NONBLOCK) != 0)
	{
	  A_LOGGING_SYSTEM_ERROR("fcntl: %s", strerror(errno));
	  close(server_fd);
	}
      else
	{
	  if (bind(server_fd, (struct sockaddr *)&serv_addr, servlen) < 0)
	    {
	      A_LOGGING_SYSTEM_ERROR("bind: %s", strerror(errno));
	      close(server_fd);
	    }
	  else
	    {
	      if (listen(server_fd, 1) < 0)
		{
		  A_LOGGING_SYSTEM_ERROR("listen: %s", strerror(errno));
		  close(server_fd);
		}
	      else
		{
		  server_watchid = a_UI_add_io_watch(xetpan, server_fd,
						     XETPAN_POLLIN |
						     XETPAN_POLLHUP |
						     XETPAN_POLLERR,	
						     Remote_accept_from_socket,
						     xetpan);
		  if (server_watchid == 0)
		    {
		      close(server_fd);
		    }
		  else
		    {
		      A_LOGGING_DEBUG("Server mode using watch id %d "
				      "for fd %d",
				      server_watchid,
				      server_fd);
		      xetpan->server_watch_id = server_watchid;
		      xetpan->server_fd = server_fd;
		      ret = TRUE;
		    }
		}
	    }
	}
    }
  return ret;
}

void
a_Remote_close_unlink_socket(XetPan *xetpan)
{
  if (xetpan->server_watch_id > 0)
    a_UI_remove_io_watch(xetpan, xetpan->server_watch_id);
  close(xetpan->server_fd);
  xetpan->server_fd = -1;  
  xetpan->server_watch_id = 0;
  unlink(xetpan->lock_socket);
}



int
a_Remote_check_for_server(XetPan *xetpan)
{
  int server_fd;
  struct sockaddr_un serv_addr;
  int servlen;

  Remote_create_sockaddr(xetpan->lock_socket, &serv_addr, &servlen);
  
  server_fd = socket(AF_UNIX, SOCK_STREAM, 0);
  if (server_fd < 0)
    {
      A_LOGGING_SYSTEM_ERROR("socket: %s", strerror(errno));
      return FALSE;
    }
  
  xetpan->server_fd = server_fd;

  if ((connect(server_fd, (struct sockaddr *) &serv_addr, servlen)) < 0)
    {
      /* That means no server is listening */
      /* In fact, this will not show up at the first invocation of
       * this function because it is invoked before command line 
       * parsing unless the user uses defines XETPAN_VERBOSE environment
       * variable.
       */
      A_LOGGING_INFO("connect: %s", strerror(errno));
      A_LOGGING_INFO(_("No running server found"));
      a_Remote_close_unlink_socket(xetpan);
      return FALSE;
    }

  return TRUE;
}


